﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;

namespace XamDockManagerTest
{
    /// <summary>
    /// Custom element used to enforce a preferred width/height while still allowing an element 
    /// to exceed the preferred size.
    /// </summary>
    public class PreferredSizeDecorator : Decorator
    {
        #region Member Variables

        private UIElement _elementBeingMeasured;

        #endregion //Member Variables

        #region Constructor
        /// <summary>
        /// Initializes a new <see cref="PreferredSizeDecorator"/>
        /// </summary>
        public PreferredSizeDecorator()
        {
        } 
        #endregion //Constructor

        #region Base class overrides

        #region ArrangeOverride
        /// <summary>
        /// Positions child elements and determines a size for this element.
        /// </summary>
        /// <param name="arrangeSize">The size available to this element for arranging its children.</param>
        /// <returns>The <see cref="System.Windows.Size"/> used by this element to arrange its children.</returns>
        protected override Size ArrangeOverride(Size arrangeSize)
        {
            UIElement element = this.Child;

            if (null != element)
            {
                // a virtualizing panel seems to determine the number of items
                // based on the measure size so we need to remeasure just in case
                this.MeasureHelper(element, arrangeSize);

                element.Arrange(new Rect(arrangeSize));
            }

            return arrangeSize;
        } 
        #endregion //ArrangeOverride

        #region MeasureOverride
        /// <summary>
        /// Invoked to measure the element and its children.
        /// </summary>
        /// <param name="availableSize">The size that reflects the available size that this element can give to its children.</param>
        /// <returns>The <see cref="System.Windows.Size"/> that represents the desired size of the element.</returns>
        protected override Size MeasureOverride(Size constraint)
        {
            UIElement element = this.Child;
            Size desiredSize = new Size();

            if (null != element)
            {
                Size measureSize = constraint;
                Size preferredSize = new Size(this.PreferredWidth, this.PreferredHeight);

                // if we're measured with an infinite value or one that exceeds the 
                // preferred then measure with the preferred constraint so we can
                // find out how large the element would be in that dimension if it
                // were only given the preferred
                if (false == double.IsNaN(preferredSize.Width))
                {
                    if (double.IsPositiveInfinity(constraint.Width) || constraint.Width > preferredSize.Width)
                        measureSize.Width = preferredSize.Width;
                }

                if (false == double.IsNaN(preferredSize.Height))
                {
                    if (double.IsPositiveInfinity(constraint.Height) || constraint.Height > preferredSize.Height)
                        measureSize.Height = preferredSize.Height;
                }

                element.Measure(measureSize);
                desiredSize = element.DesiredSize;
            }

            return desiredSize;
        }
        #endregion //MeasureOverride

        #region OnChildDesiredSizeChanged
        /// <summary>
        /// Invoked when the desired size of a child has been changed.
        /// </summary>
        /// <param name="child">The child whose size is being changed.</param>
        protected override void OnChildDesiredSizeChanged(UIElement child)
        {
            if (child == _elementBeingMeasured)
                return;

            base.OnChildDesiredSizeChanged(child);
        } 
        #endregion //OnChildDesiredSizeChanged

        #endregion //Base class overrides

        #region Properties

        #region PreferredWidth

        /// <summary>
        /// Identifies the <see cref="PreferredWidth"/> dependency property
        /// </summary>
        public static readonly DependencyProperty PreferredWidthProperty = DependencyProperty.Register("PreferredWidth",
            typeof(double), typeof(PreferredSizeDecorator), new FrameworkPropertyMetadata(double.NaN));

        /// <summary>
        /// Returns the preferred width for the element
        /// </summary>
        /// <seealso cref="PreferredWidthProperty"/>
        public double PreferredWidth
        {
            get
            {
                return (double)this.GetValue(PreferredSizeDecorator.PreferredWidthProperty);
            }
            set
            {
                this.SetValue(PreferredSizeDecorator.PreferredWidthProperty, value);
            }
        }

        #endregion //PreferredWidth

        #region PreferredHeight

        /// <summary>
        /// Identifies the <see cref="PreferredHeight"/> dependency property
        /// </summary>
        public static readonly DependencyProperty PreferredHeightProperty = DependencyProperty.Register("PreferredHeight",
            typeof(double), typeof(PreferredSizeDecorator), new FrameworkPropertyMetadata(double.NaN));

        /// <summary>
        /// Returns the preferred width for the element
        /// </summary>
        /// <seealso cref="PreferredHeightProperty"/>
        public double PreferredHeight
        {
            get
            {
                return (double)this.GetValue(PreferredSizeDecorator.PreferredHeightProperty);
            }
            set
            {
                this.SetValue(PreferredSizeDecorator.PreferredHeightProperty, value);
            }
        }

        #endregion //PreferredHeight

        #endregion //Properties

        #region Methods
        
        #region MeasureHelper
        private void MeasureHelper(UIElement element, Size measureSize)
        {
            UIElement oldElement = _elementBeingMeasured;
            _elementBeingMeasured = element;
            element.Measure(measureSize);
            _elementBeingMeasured = oldElement;
        }
        #endregion //MeasureHelper

        #endregion //Methods
    }
}
